<?php

namespace Drupal\helpdesk_integration;

use Drupal\comment\Entity\Comment;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Plugin\PluginBase as CorePluginBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Http\ClientFactory;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\user\Entity\User;
use Drupal\user\UserDataInterface;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for helpdesk_integration plugins.
 */
abstract class PluginBase extends CorePluginBase implements PluginInterface, ContainerFactoryPluginInterface {

  /**
   * TBD.
   *
   * @var \Drupal\Core\Http\ClientFactory
   */
  protected $clientFactory;

  /**
   * TBD.
   *
   * @var \Drupal\helpdesk_integration\Service
   */
  protected $service;

  /**
   * TBD.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * TBD.
   *
   * @var \Drupal\user\UserDataInterface
   */
  protected $userData;

  /**
   * TBD.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * TBD.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * TBD.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * PluginBase constructor.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Http\ClientFactory $client_factory
   *   HTTP client factory.
   * @param \Drupal\helpdesk_integration\Service $service
   *   Helpdesk services.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager service.
   * @param \Drupal\user\UserDataInterface $user_data
   *   User data service to persistantly store data for each user.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   Drupal's time service.
   * @param \Psr\Log\LoggerInterface $logger
   *   Drupal's logger service.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The currently logged in user account session.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, ClientFactory $client_factory, Service $service, EntityTypeManagerInterface $entity_type_manager, UserDataInterface $user_data, TimeInterface $time, LoggerInterface $logger, AccountInterface $current_user) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->clientFactory = $client_factory;
    $this->service = $service;
    $this->entityTypeManager = $entity_type_manager;
    $this->userData = $user_data;
    $this->time = $time;
    $this->logger = $logger;
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('http_client_factory'),
      $container->get('helpdesk_integration.service'),
      $container->get('entity_type.manager'),
      $container->get('user.data'),
      $container->get('datetime.time'),
      $container->get('logger.factory')->get('helpdesk'),
      $container->get('current_user')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function label(): string {
    return (string) $this->pluginDefinition['label'];
  }

  /**
   * {@inheritdoc}
   */
  public function findUserByData(HelpdeskInterface $helpdesk, $key, $data): ?UserInterface {
    static $users = [];
    $hash = is_scalar($data) ? $data : hash('md5', $data);
    if (empty($users[$helpdesk->id()][$key][$hash])) {
      $load_uid = 1;
      $data_key = implode(' ', [$helpdesk->id(), $key]);
      foreach ($this->userData->get('helpdesk_integration', NULL, $data_key) as $uid => $value) {
        if ($value === $data) {
          $load_uid = $uid;
          break;
        }
      }
      /** @var \Drupal\user\UserInterface $user */
      $users[$helpdesk->id()][$key][$hash] = User::load($load_uid);
    }
    return $users[$helpdesk->id()][$key][$hash];
  }

  /**
   * {@inheritdoc}
   */
  public function getUserData(UserInterface $user, HelpdeskInterface $helpdesk, $key) {
    $data_key = implode(' ', [$helpdesk->id(), $key]);
    return $this->userData->get('helpdesk_integration', $user->id(), $data_key);
  }

  /**
   * {@inheritdoc}
   */
  public function setUserData(UserInterface $user, HelpdeskInterface $helpdesk, $key, $value): void {
    $data_key = implode(' ', [$helpdesk->id(), $key]);
    $this->userData->set('helpdesk_integration', $user->id(), $data_key, $value);
  }

  /**
   * {@inheritdoc}
   */
  final public function updateAllIssues(HelpdeskInterface $helpdesk, UserInterface $user = NULL, $force = FALSE): void {
    if ($user === NULL) {
      $user = User::load($this->currentUser->id());
    }
    $last_sync = $this->getUserData($user, $helpdesk, 'last_sync');
    $current = $this->time->getRequestTime();
    if (!$force && ($last_sync + self::REFRESH_INTERVAL >= $current)) {
      return;
    }
    $this->setUserData($user, $helpdesk, 'last_sync', $current);

    $issueStorage = $this->entityTypeManager->getStorage('helpdesk_issue');
    $commentStorage = $this->entityTypeManager->getStorage('comment');
    $remoteIssueIds = [];
    foreach ($this->getAllIssues($helpdesk, $user) as $issue) {
      $changed = FALSE;
      /** @var \Drupal\helpdesk_integration\IssueInterface $localIssue */
      if ($localIssues = $issueStorage->loadByProperties(['helpdesk' => $helpdesk->id(), 'extid' => $issue->get('extid')->value])) {
        $localIssue = array_shift($localIssues);
        // Update issue fields if necessary.
        foreach (['resolved', 'title', 'status', 'changed', 'body'] as $field_name) {
          if ($localIssue->get($field_name)->value !== $issue->get($field_name)->value) {
            $localIssue->set($field_name, $issue->get($field_name)->value);
            $changed = TRUE;
          }
        }
      }
      else {
        $localIssue = $issue;
        $changed = TRUE;
      }
      if ($localIssue->addUser($user->id()) || $changed) {
        $localIssue->save();
      }
      $remoteIssueIds[] = $localIssue->id();

      // Sync comments.
      $remoteCommentIds = [];
      foreach ($issue->getComments() as $comment) {
        $changed = FALSE;
        /** @var \Drupal\Comment\CommentInterface $localComment */
        if ($localComments = $commentStorage->loadByProperties([
          'entity_type' => 'helpdesk_issue',
          'entity_id' => $localIssue->id(),
          'field_extid' => $comment['extid'],
        ])) {
          $localComment = array_shift($localComments);
          // Update comment fields if necessary.
          foreach ([
            'uid' => 'author',
            'comment_body' => 'body',
            'created' => 'created',
            'changed' => 'changed',
          ] as $local_field_name => $remote_field_name) {
            if ($localComment->get($local_field_name)->value !== $comment[$remote_field_name]) {
              $localComment->set($local_field_name, $comment[$remote_field_name]);
              $changed = TRUE;
            }
          }
        }
        else {
          $localComment = Comment::create([
            'entity_type' => 'helpdesk_issue',
            'entity_id' => $localIssue->id(),
            'comment_type' => 'helpdesk_issue_comment',
            'field_name' => 'comment',
            'field_extid' => $comment['extid'],
            'uid' => $comment['author']->id(),
            'created' => $comment['created'],
            'changed' => $comment['changed'],
            'comment_body' => [
              'value' => $comment['body'],
              'format' => 'basic_html',
            ],
          ]);
          $changed = TRUE;
        }
        if ($changed) {
          $localComment->save();
        }
        $remoteCommentIds[] = $localComment->id();

        // Remove existing local comments that no longer exist remotely.
        $localComments = $commentStorage->loadByProperties([
          'entity_type' => 'helpdesk_issue',
          'entity_id' => $localIssue->id(),
        ]);
        $removeCommentIds = array_diff(array_keys($localComments), $remoteCommentIds);
        /** @var \Drupal\Comment\CommentInterface $removeComment */
        foreach ($removeCommentIds as $removeCommentId) {
          $localComments[$removeCommentId]->delete();
        }

        // TODO: Handle attachments, see drupalspoons/helpdesk_integration#17.
      }
    }

    // Remove access to existing items that no longer apply.
    $localIssues = $issueStorage->loadByProperties(['helpdesk' => $helpdesk->id(), 'users' => $user->id()]);
    $removeIssueIds = array_diff(array_keys($localIssues), $remoteIssueIds);
    /** @var \Drupal\helpdesk_integration\IssueInterface $removeIssue */
    foreach ($removeIssueIds as $removeIssueId) {
      $removeIssue = $localIssues[$removeIssueId];
      if ($removeIssue->removeUser($user->id())) {
        $removeIssue->save();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  final public function updateAllUsers(HelpdeskInterface $helpdesk, UserInterface $user = NULL): void {
    $users = [];
    if ($user === NULL) {
      /** @var \Drupal\user\UserInterface $user */
      foreach (User::loadMultiple() as $user) {
        if ($helpdesk->hasAccess($user)) {
          $users[] = $user;
        }
      }
    }
    else {
      $users[] = $user;
    }

    try {
      foreach ($users as $user) {
        $this->pushUser($helpdesk, $user);
      }
    }
    catch (HelpdeskPluginException $e) {
      $this->logger->alert('Pushing users to @helpdesk failed: @exception', [
        '@helpdesk' => $helpdesk->label(),
        '@exception' => $e->getMessage(),
      ]);
    }
  }

}
